#ifndef ROSE_NODES_H_INCLUDED_
#define ROSE_NODES_H_INCLUDED_

#include "rose.h"
#include "../internal_representation/design.h"

namespace risc {

extern SgProject* get_project();
};

namespace risc {

namespace tools {

class RoseNodes {

public:

  static const std::vector<SgNode*>& get_func_defs()
  {
    initialize();
    return func_defs_;
  }

  static const std::vector<SgNode*>& get_func_decls()
  {
    initialize();
    return func_decls_;
  }

  static const std::vector<SgNode*>& get_class_defs()
  {
    initialize();
    return class_defs_;
  }

  static const std::vector<SgNode*>& get_class_decls()
  {
    initialize();
    return class_decls_;
  }

  static const std::vector<SgNode*>& get_namespace_defs()
  {
    initialize();
    return namespace_defs_;
  }

  static const std::vector<SgNode*>& get_template_instance_decl()
  {
    initialize();
    return template_instance_decl_;
  }

  static const std::vector<SgNode*>& get_typedef_decl()
  {
    initialize();
    return typedef_decl_;
  }

private:

  static bool is_initialized_;

  static std::vector<SgNode*> func_defs_;
  static std::vector<SgNode*> func_decls_;
  static std::vector<SgNode*> class_defs_;
  static std::vector<SgNode*> class_decls_;
  static std::vector<SgNode*> namespace_defs_;
  static std::vector<SgNode*> template_instance_decl_;
  static std::vector<SgNode*> typedef_decl_;

  static void initialize()
  {
    if(RoseNodes::is_initialized_ == true) {
      return;
    }

    SgProject *project = risc::get_project();

    assert(project && "Project is not yet initialized");

    RoseNodes::func_defs_
      = NodeQuery::querySubTree(project, V_SgFunctionDefinition);

    RoseNodes::func_decls_
      = NodeQuery::querySubTree(project, V_SgFunctionDeclaration);

    RoseNodes::class_defs_
      = NodeQuery::querySubTree(project, V_SgClassDefinition);

    RoseNodes::class_decls_
      = NodeQuery::querySubTree(project, V_SgClassDeclaration);

    RoseNodes::namespace_defs_
      = NodeQuery::querySubTree(project, V_SgNamespaceDefinitionStatement);

    RoseNodes::template_instance_decl_
      = NodeQuery::querySubTree(project, V_SgTemplateInstantiationDecl);

    RoseNodes::typedef_decl_
      = NodeQuery::querySubTree(project, V_SgTypedefDeclaration);

    RoseNodes::is_initialized_ = true;
  }
};

}; // end of namespace tools

}; // end of namespace risc

#endif /* ROSE_NODES_H_INCLUDED_ */

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
