#include "systemc.h"

class Tile1: public sc_module {

public:

  SC_HAS_PROCESS(Tile1);

  int &reference_;
  int variable_;

  Tile1(sc_module_name mn, int &ref):
    sc_module(mn),
    reference_(ref)
  {
    SC_THREAD(func);
  }

  void func()
  {
    reference_ = 42;
//    wait();
  }
};

class Top1: public sc_module {

public:

  SC_HAS_PROCESS(Top1);

  int top_reference0;
  int top_reference1;
  int top_reference2;
  int top_reference3;

  Tile1 *tiles_[4][4];

  Top1(sc_module_name mn):
    sc_module(mn)
  {
    // block 0
    tiles_[0][0] = new Tile1("tile00", top_reference0);
    tiles_[0][1] = new Tile1("tile01", top_reference0);
    tiles_[1][0] = new Tile1("tile10", top_reference0);
    tiles_[1][1] = new Tile1("tile11", top_reference0);

    // block 1
    tiles_[0][2] = new Tile1("tile02", top_reference1);
    tiles_[0][3] = new Tile1("tile03", top_reference1);
    tiles_[1][2] = new Tile1("tile12", top_reference1);
    tiles_[1][3] = new Tile1("tile13", top_reference1);

    // block 2
    tiles_[2][0] = new Tile1("tile20", top_reference2);
    tiles_[2][1] = new Tile1("tile21", top_reference2);
    tiles_[3][0] = new Tile1("tile30", top_reference2);
    tiles_[3][1] = new Tile1("tile31", top_reference2);

    // block 3
    tiles_[2][2] = new Tile1("tile22", top_reference3);
    tiles_[2][3] = new Tile1("tile23", top_reference3);
    tiles_[3][2] = new Tile1("tile32", top_reference3);
    tiles_[3][3] = new Tile1("tile33", top_reference3);
  }
};

class Tile2: public sc_module {

public:

  SC_HAS_PROCESS(Tile2);

  int &reference_;

  Tile2(sc_module_name mn, int &ref):
    sc_module(mn),
    reference_(ref)
  {
    SC_THREAD(func);
  }

  void func()
  {
    reference_ = 42;
//    wait();
  }
};

int global_variable;
int global_variable1;
int global_variable2;

class Top2: public sc_module {

public:

  SC_HAS_PROCESS(Top2);

  int top_reference0;
  int top_reference1;
  int top_reference2;
  int top_reference3;
  int some_variable_;

  Tile2 *tiles_[2][2];

  Top2(sc_module_name mn):
    sc_module(mn)
  {
    // block 0
    tiles_[0][0] = new Tile2("tile00", top_reference0);
    tiles_[0][1] = new Tile2("tile01", top_reference1);
    tiles_[1][0] = new Tile2("tile10", top_reference2);
    tiles_[1][1] = new Tile2("tile11", top_reference3);

    SC_THREAD(thread);
  }

  void thread()
  {
    top_reference0 = 1;
  }
};

class SenderTile: public sc_module {

  SC_HAS_PROCESS(SenderTile);

public:

  sc_event &event_ref_;
  int &x_ref_;

  SenderTile(sc_module_name mn, sc_event &event, int &x):
    sc_module(mn),
    event_ref_(event),
    x_ref_(x)
  {
    SC_THREAD(thread);
    SC_THREAD(thread1);
  }

  void thread()
  {
    event_ref_.notify(SC_ZERO_TIME);
  }

  void thread1()
  {
    x_ref_;
  }
};

class ReceiverTile: public sc_module {

  SC_HAS_PROCESS(ReceiverTile);

public:

  sc_event &event_;
  int &x_ref_;

  ReceiverTile(sc_module_name mn, sc_event &event, int &x):
    sc_module(mn),
    event_(event),
    x_ref_(x)
  {
    SC_THREAD(thread);
    SC_THREAD(thread1);
  }

  void thread()
  {
    wait(event_);
  }

  void thread1()
  {
    x_ref_ = 5;
  }
};

class Unit: public sc_module {

  SC_HAS_PROCESS(Unit);

public:

  SenderTile sender_;
  ReceiverTile receiver_;

  int x;
  sc_event event_;

  Unit(sc_module_name mn, sc_event &event):
    sc_module(mn),
    sender_("sender", event_, x),
    receiver_("recevier", event_, x)
  {
  }
};


class Grid: sc_module {

  SC_HAS_PROCESS(Grid);

public:

  sc_event *events_;
  Unit ***units_;

  int rows_;
  int columns_;

  Grid(sc_module_name mn, int rows, int columns):
    rows_(rows),
    columns_(columns)
  {
    events_ = new sc_event[rows * columns];
    units_ = new Unit**[rows];

    for(int i=0; i<rows; i++) {
      units_[i] = new Unit* [columns];
    }

    for(int x=0, counter=0; x<rows; x++) {
      for(int y=0; y<columns; y++) {

        std::stringstream ss;
        ss << counter;
        std::string name = "name_";
        std::string temp = ss.str();
        name += temp;

        units_[x][y] = new Unit(name.c_str(), events_[counter]);
        counter++;
      }
    }
  }
};

class SubChannelSender1: public sc_module {

  SC_HAS_PROCESS(SubChannelSender1);

public:
  int &ref_;
  sc_event &event_;

  SubChannelSender1(sc_module_name mn, sc_event &event, int &ref):
    sc_module(mn),
    ref_(ref),
    event_(event)
  {
    SC_THREAD(thread1);

    SC_THREAD(thread2);
  }

  void thread1()
  {
    ref_;
  }

  void thread2()
  {
    event_.notify(SC_ZERO_TIME);
  }
};

class SubModuleReceiver1: public sc_module {

  SC_HAS_PROCESS(SubModuleReceiver1);

public:
  int &ref_;
  sc_event &event_;

  SubModuleReceiver1(sc_module_name mn, sc_event &event, int &ref):
    sc_module(mn),
    event_(event),
    ref_(ref)
  {
    SC_THREAD(thread1);

    SC_THREAD(thread2);
  }

  void thread1()
  {
    ref_ = 5;
  }

  void thread2()
  {
    wait(event_);
  }
};

class TopModule1: public sc_module {

public:
  int var_0;
  int var_1;
  int var_2;
  int var_3;

  sc_event event_0;
  sc_event event_1;
  sc_event event_2;
  sc_event event_3;

  SubChannelSender1 *sub_channel[4];
  SubModuleReceiver1 *sub_module[4];

  TopModule1(sc_module_name mn):
    sc_module(mn)
  {
    sub_channel[0] = new SubChannelSender1("SubChannelSender10", event_0, var_0);
    sub_channel[1] = new SubChannelSender1("SubChannelSender11", event_1, var_1);
    sub_channel[2] = new SubChannelSender1("SubChannelSender12", event_2, var_2);
    sub_channel[3] = new SubChannelSender1("SubChannelSender13", event_3, var_3);

    sub_module[0] = new SubModuleReceiver1("SubModuleReceiver10", event_0, var_0);
    sub_module[1] = new SubModuleReceiver1("SubModuleReceiver11", event_1, var_1);
    sub_module[2] = new SubModuleReceiver1("SubModuleReceiver12", event_2, var_2);
    sub_module[3] = new SubModuleReceiver1("SubModuleReceiver13", event_3, var_3);
  }
};

class TopChannel2: public sc_channel {

public:
  int var_0;
  int var_1;
  int var_2;
  int var_3;

  sc_event event_0;
  sc_event event_1;
  sc_event event_2;
  sc_event event_3;

  SubChannelSender1 *sub_channel[4];
  SubModuleReceiver1 *sub_module[4];

  TopChannel2(sc_module_name mn):
    sc_channel(mn)
  {
    sub_channel[0] = new SubChannelSender1("SubChannelSender10", event_0, var_0);
    sub_channel[1] = new SubChannelSender1("SubChannelSender11", event_1, var_1);
    sub_channel[2] = new SubChannelSender1("SubChannelSender12", event_2, var_2);
    sub_channel[3] = new SubChannelSender1("SubChannelSender13", event_3, var_3);

    sub_module[0] = new SubModuleReceiver1("SubModuleReceiver10", event_0, var_0);
    sub_module[1] = new SubModuleReceiver1("SubModuleReceiver11", event_1, var_1);
    sub_module[2] = new SubModuleReceiver1("SubModuleReceiver12", event_2, var_2);
    sub_module[3] = new SubModuleReceiver1("SubModuleReceiver13", event_3, var_3);
  }
};

class TopChannelTestModule: public sc_module {

public:

  TopModule1 tm1;
  TopChannel2 tc1;

  TopChannelTestModule(sc_module_name mn):
    sc_module(mn),
    tm1("tm1"),
    tc1("cm1")
  { }
};

class Top: public sc_module {

public:

  Top1 t1; // grid of tiles with mapped references
  Top2 t2; // gird of tiles with mapped references,
           // hosting module of grid accesses one reference as well
  Grid grid; // grid of units. A unit has a sender and receiver which
             // have a data conflict and a notification dependency

  TopChannelTestModule tct; // first module and later mixed up with channels
                            // in the hierarchy

  Top(sc_module_name mn):
    t1("top1"),
    t2("top2"),
    grid("grid", 4, 4),
    tct("TopChannelTest")
  { }
};


int sc_main(int argc, char **argv)
{
  Top t("top");
  sc_start();

  return 0;
}
